#include <stdio.h>
#include <fcntl.h>
#include <sys/stat.h>
#include <sys/file.h>
#include <syslog.h>
#include <string.h>
#include <time.h>
#include <ctype.h>
#include <unistd.h>
#include <cost.h>
#include <strfunc.h>

#ifdef	ACCOUNTING_SERVICE
#include <desire.h>
#endif

#include <lgba.h>

#ifdef AEMON_MODE
#include <stdlib.h>
#endif

#ifndef POS_FILE
#define POS_FILE	"/adm/tac/logs/acct.pos"
#endif

#ifndef ACCTER_PID_FILE
#define ACCTER_PID_FILE	"/var/run/accter.pid"
#endif

#ifndef CONFIG
#define CONFIG	"/adm/tac/conf/utmpacct.conf"
#endif

char mybuf[16384];

int detect(char *);
int decrease(char *, char *, char *, char *, char *, char *);

#ifdef ACCT_DIR
#include "acct.h"

/*
int gethashedfd(char *, time_t *);
typedef struct _FH_ {
	int fd;
	char *user;
	int year;
	time_t lasttime;
} FDHASH;
#define FDHASH_SIZE 32
FDHASH fdhash[FDHASH_SIZE];
*/

#endif

char *account;
char *acctservice;

int main() {
	int fd, les;
	int posfd;
	size_t size;
	struct stat sb;
	char **vp;
	off_t ofs=0;
	ino_t ino=0;
	int cycle;
	int pf;
	char *s;
	slist *sl;

	if(cfgread(CONFIG)) exit(2);

	if(!(account=cfgget("account"))) {
		printf("Account attribute definition not found in config.\n");
		exit(2);
	};

	if(!(s=cfgget("acctfile")) || (fd=open(s, O_RDONLY)) == -1) {
		fprintf(stderr, "Unable to open an accounting file: %s.\n", s);
		exit(2);
	};

	if((pf=open(ACCTER_PID_FILE, O_WRONLY | O_CREAT, 0644)) == -1) {
		fprintf(stderr, "Error: can't open pid file: %s.\n", ACCTER_PID_FILE);
		exit(2);
	};
	if(flock(pf, LOCK_EX | LOCK_NB)) {
		close(pf);
		fprintf(stderr, "Error: previous copy already running.\n");
		exit(2);
	};

	if((posfd=open(POS_FILE, O_RDWR | O_CREAT, 0644)) == -1 ||
	fstat(posfd, &sb)) {
		fprintf(stderr, "Unable to open position file.\n");
		exit(2);
	};

	if(sb.st_size != sizeof(off_t) + sizeof(ino_t)) {
		fstat(fd, &sb);
		ino=sb.st_ino;
		write(posfd, &ino, sizeof(ino_t));
		write(posfd, &ofs, sizeof(off_t));
	} else {
		read(posfd, &ino, sizeof(ino_t));
		read(posfd, &ofs, sizeof(off_t));
		fstat(fd, &sb);
		if(ino != sb.st_ino) {
			ofs=0;
			ino=sb.st_ino;
			lseek(posfd, 0, SEEK_SET);
			write(posfd, &ino, sizeof(ino_t));
			write(posfd, &ofs, sizeof(off_t));
		} else if(ofs > sb.st_size) {
			ofs=0;
			lseek(posfd, 0, SEEK_SET);
			write(posfd, &ino, sizeof(ino_t));
			write(posfd, &ofs, sizeof(off_t));
		};
	}

	lseek(fd, ofs, SEEK_SET);

	if(lc_init(NULL)) {
		fprintf(stderr, "libcost init failed\n");
		exit(2);
	};

	if(lgba_init(NULL) || lgba_open()) {
		lgba_perror("Error");
		exit(2);
	};

	lgba_setuser("root");
#ifdef LGBA_NODE
	if((les=chnode(LGBA_NODE)))
		fprintf(stderr, "Error in chnode(): %s.\n", lgba_err2string(les));
#endif

#ifdef	ACCOUNTING_SERVICE
	if(!(acctservice = cfgget("acctservice")))
		acctservice=ACCOUNTING_SERVICE;
	if(ctb_session(NULL, acctservice, NULL)) {
		fprintf(stderr, "Unable to open CTB session.\n");
		exit(2);
	};
#endif

#ifdef AEMON_MODE
	fprintf(stderr, "ACCTer: All OK. Daemonizing.\n");
	daemon(0, 0);
#endif
	/* saving PID */
	les=strlen(s=genstr(getpid()))+1;
	s[les-1]='\n';
	write(pf, s, les);
	ftruncate(pf, les);
	fsync(pf);

	openlog("accter", LOG_CONS, LOG_LOCAL6);
	syslog(LOG_INFO, "ACCT file control facility started: %d.", getpid());

	sl=sinit(NULL);

#ifdef ACCT_DIR
	memset(&fdhash, 0, sizeof(fdhash));
#endif

	cycle = 0;
	while(1) {
		while(( size=read(fd, mybuf, sizeof(mybuf)) ) > 0){
			s=mybuf+size-1; /* to the last byte of buffer */
			cycle=1;

			/* Avoid improperly handled strings */
			if(	*s != '\n') {
				while(--s > mybuf)
					if(*s == '\n') break;

				/* if whole buffer does not contain '\n' at all */
				if(s == mybuf) {
					lseek(fd, ofs, SEEK_SET);
					continue;
				} else {
					size=(s - mybuf);
					lseek(fd, ofs + size, SEEK_SET);
				};
			};

			*s = 0; /* Clear the terminating '\n' */
			ofs+=size;
			sclear(sl);
			split4(sl, mybuf, "\n");
			vp=sarray(sl, 0);
			vp--;
			while(*(++vp))
				detect(*vp);
		}; /* while(read) */

		if(cycle) {
			cycle=0;
			lseek(posfd, sizeof(ino), SEEK_SET);
			write(posfd, &ofs, sizeof(off_t));
			fsync(posfd);
			/*
			for(les=0; les<FDHASH_SIZE; les++)
				if(fdhash[les].user) fsync(fdhash[les].fd);
			*/
			continue;
		}
		usleep(250000);
	}

}

int detect(char *str) {
#ifdef PSEUDO_STATIC_IP
	int les;
	char *ipaddr;
#endif
	static slist *sl=NULL;
	static slist *at=NULL;
	static slist *va=NULL;
	char **vp;
	char *p;

	if(sl) {
		sclear(sl); sclear(at); sclear(va);
	} else {
		sl=sinit(NULL); at=sinit(NULL); va=sinit(NULL);
	};

	split4(sl, str, " \t");
	if(sl->count < 8) return 0;

	vp=sarray(sl, 8) - 1;
	while(*(++vp)) {
		if((p=strchr(*vp, '='))) *p++=0;
		else p="";
		sadd(at, *vp);
		sadd(va, p);
	};
#	define TME	(sl->list[2])
#	define USER	(sl->list[4])
#	define LINE	(sl->list[5])
#	define SUS	(sl->list[7])	/* start, update, stop... */

	if(eq(SUS, "stop"))
		decrease(TME, USER,
			sget2(at, "elapsed_time", va),
			sget2(at, "bytes_in", va),
			sget2(at, "bytes_out", va),
			LINE);

#ifdef PSEUDO_STATIC_IP
	if(!(ipaddr=sget2(at, "address", va)) &&
	!(ipaddr=sget2(at, "addr", va))) return 0;

	{
		static char *ipnode=NULL, *usenode=NULL;
		if(!ipnode) {
			ipnode=cfgget("ipnode");
			if(!ipnode) ipnode=IP_NODE;
		};
		if(!usenode) usenode=cfgget("node");

	if((les=lgba_chnode(ipnode)) ||
		(les=load(ipaddr)) ||
		(les=lgba_chnode(usenode))) {

		lgba_chnode(usenode);
		syslog(LOG_ERR, "Load IP object failed: %s @ node '%s'.\n", lgba_err2string(les), ipnode);
		return 0;
	};
	};

	if(eq(SUS, "start") || eq(SUS, "update")) {
		setattr("user", USER);
		setattr("status", "inuse");
	} else if(eq(SUS, "stop")) {
		setattr("status", "free");
	};

	lgba_saveloaded(1);
#endif /* PSEUDO_STATIC_IP */

	return 0;
};

#ifdef	ACCOUNTING_SERVICE
	ctb_data	data;
#endif

int decrease(char *tme, char *user, char *ela, char *bi, char *bo, char *line) {
	int les;
	int cost;
	time_t start;
	unsigned int elapsed;

	if(!bi) bi="0";
	if(!bo) bo="0";
	if(!ela) ela="0";

	if((les=load(user))) {
		syslog(LOG_ERR, "Error loading object: %s.\n", lgba_err2string(les));
		return les;
	};

	elapsed=(unsigned int)atoi(ela);
	start=(time_t)atoi(tme)-elapsed;
	cost=lc_timecost(start, elapsed, "online",
		getattr2("objectclass"));

	{	/* Lower-case the name */
		char *tc=user;
		while(*tc) *tc++=tolower(*tc);
	};

#ifdef EXTERN
	sprintf(mybuf, EXTERN, user, (int)cost);
	system(mybuf);
#endif

	syslog(LOG_INFO, "Decreasing time for %s: time spent %s, cost %d, time left: %d.", user, ela, cost, atoi(getattr(account)) - cost);

	{
	char buf[64];

	if(cost) sprintf(buf, "%+d", -cost);
	else sprintf(buf, "=%d", atoi(getattr(account)));

	setattr(account, buf);
	};

	lgba_saveloaded(1);


#ifdef	ACCOUNTING_SERVICE
	if(!acctservice || !*acctservice) return 0;

	data.TS=(time_t)atoi(tme);
	data.FL.s.s1=0;
	data.FL.s.s2=(unsigned short)atoi(line + strcspn(line, "0123456789"));
	data.In=(unsigned int)atoi(bo);
	data.Out=(unsigned int)atoi(bi);
	data.Src=cost;
	data.Dst=0;
	data.U1=0;
	data.U2=elapsed;
	if(!load(user)) (int)data.Dst=atoi(getattr(account));
	if(ctb_chname(user))
		syslog(LOG_ERR, "Can't change name to '%s'", user);
	else if(ctb_add(&data))
		syslog(LOG_ERR, "Can't append CTB log for '%s'", user);
	ctb_chname("");	/* close opened in ctb_add() file descriptor */

#endif

	return 0;
};

/*
#ifdef ACCOUNTING_SERVICE

int gethashedfd(char *user, time_t *timep) {
	int year;
	int i;
	char *tl;
	time_t min=fdhash[0].lasttime;
	FDHASH *fdp=fdhash;
	FDHASH *minfdp=fdhash;

	year=(localtime(timep))->tm_year+1900;

	for(i=0; i<FDHASH_SIZE; i++, fdp++) {
		if(fdp->lasttime < min) min=(minfdp=fdp)->lasttime;
		if(!fdp->user || fdp->year != year) continue;
		if(strcasecmp(fdp->user, user)) continue;
		return fdp->fd;
	};

	fdp=minfdp;

	snprintf(mybuf, sizeof(mybuf), "%s/%s~%d", acctdir, user, year);
	tl=mybuf + strlen(acctdir);
	while(*tl) *tl++=tolower(*tl);
	if((i=open(mybuf, O_WRONLY | O_APPEND | O_CREAT, 0644)) == -1) return -1;

	if(fdp->user) {
		free(fdp->user);
		close(fdp->fd);
	};
	fdp->user=strdup(user);
	fdp->fd=i;
	fdp->year=year;
	fdp->lasttime=*timep;
	return i;
};

#endif
*/
