#include <stdio.h>
#include <string.h>
#include <sys/types.h>
#include <lber.h>
#include <ldap.h>
#include <lgba.h>
#include <strfunc.h>
#include <syslog.h>
#include <sys/file.h>
#include <stdlib.h>
#include "tac_plus.h"
#include "utmp.h"

char *u;
char buffer[256];
int protocol_ip=0;

int countonlineuser(char *name);

#ifndef PSEUDO_STATIC_IP
#error Sorry, you must define PSEUDO_STATIC_IP to compile this program.
#endif

int getipfunc(int);

#ifndef STATIC_IP_ATTRIBUTE
#define STATIC_IP_ATTRIBUTE	"StaticIPAddress"
#endif

#ifndef AVP_IP_ATTRIBUTE
#define AVP_IP_ATTRIBUTE "avpair_ip"
#endif

char *usenode, *maxsessions, *ipnode;

int main(int ac, char **av) {
	char *filter;
	int fd;

	if(ac < 2) { printf("Unknown user\n"); exit(1); };
	u=av[1];

	while(fgets(buffer, sizeof(buffer), stdin)) {
		printf("%s", buffer);
		if(!strcmp(buffer, "protocol=ip\n"))
			protocol_ip=1;
	};

#ifndef	MAXSESSIONS
	if(!protocol_ip) exit(0);
#endif

	openlog("setavp", LOG_CONS, LOG_LOCAL6);

	if(cfgread(CONFIG)) {
		syslog(LOG_ERR, "Error: Configuration error!");
		exit(1);
	};

	usenode=cfgget("node");
	if(!(ipnode=cfgget("ipnode")))
		ipnode=IP_NODE;
	if(!(maxsessions=cfgget("maxsessions")))
		maxsessions="maxsessions";

	if(lgba_init(NULL) || lgba_open()) {
		syslog(LOG_ERR, "Error: lgba_init() error!");
		exit(1);
	};

	if(lgba_chnode(usenode)) {
		syslog(LOG_ERR, "Error: lgba_chnode(\"%s\") error!", usenode);
		exit(1);
	};

	{
#ifdef MAXSESSIONS
char *uatl[]={ STATIC_IP_ATTRIBUTE, AVP_IP_ATTRIBUTE, maxsessions, NULL };
#else
	char *uatl[]={ STATIC_IP_ATTRIBUTE, AVP_IP_ATTRIBUTE, NULL };
#endif
	lgba_setattrlist(uatl);
	};

	if(load(u) == 0) {
		char **ip;
		char **avs;
		char *s;
		int ms;
		int co;

#ifdef	MAXSESSIONS
		s=getattr(maxsessions);
		ms=(*s)?atoi(s):1;
		if(ms <= 0 || ms > MAXSESSIONS) ms=MAXSESSIONS;
		if((co=countonlineuser(u)) > ms) {
			printf("Maximum sessions limited to %d (%d already online)\n", ms, co);
			exit(1);
		};
		if(!protocol_ip && co == ms) {
			printf("Maximum sessions limited to %d (%d already online)\n", ms, co);
			exit(1);
		};
		if(!protocol_ip) exit(0);
#endif

#ifdef AVP_IP_ATTRIBUTE
		if((avs=getattr2(AVP_IP_ATTRIBUTE)) && *avs)
			printf("%s\n", join(avs, "\n"));
#endif

#ifdef STATIC_IP_ATTRIBUTE
		if((ip=getattr2(STATIC_IP_ATTRIBUTE)) && *ip) {
			printf("addr=%s\n", *ip);
			syslog(LOG_INFO, "%s got static ip: %s", u, *ip);
			lgba_close();
			exit(2);
		};
#endif
	};

#ifndef MAXSESSIONS
	if(!protocol_ip) exit(0);
#endif

	if(lgba_chnode(ipnode)) {
		printf("LGBA lgba_chnode(\"%s\") Error\n", ipnode);
		exit(1); }

	filter=(char *)malloc(sizeof("(&(status=leased)(user=smb))") + strlen(u));

	/* simple locking */
	if((fd=open(*av, O_WRONLY)) > -1)
		flock(fd, LOCK_EX);

	/* Get free address which belongs to this user */

	sprintf(filter, "(&(status=free)(user=%s))", u);
	lgba_search(NULL, filter, NULL, getmodifiedattributename(),
		NULL, NULL, getipfunc, 0);

	/* Get leasing address which belongs to this user.
	   BUGS:
		If two users with the same login names are trying
		to connect at the same time, there is possible to give
		them two equal IP addresses. But it is not a common.
	*/

	sprintf(filter, "(&(status=leasing)(user=%s))", u);
	lgba_search(NULL, filter, NULL, getmodifiedattributename(),
		NULL, NULL, getipfunc, 0);

	/* Get any free address */

	lgba_search(NULL, "status=free", NULL, getmodifiedattributename(),
		NULL, NULL, getipfunc, 0);

	syslog(LOG_WARNING, "Warning: No free address found.");

	lgba_search(NULL, "(!(status=inuse))", NULL, getmodifiedattributename(),
		NULL, NULL, getipfunc, 0);

	syslog(LOG_WARNING, "Error: no available unused addresses for %s!", u);

	sprintf(filter, "(user=%s)", u);
	lgba_search(NULL, filter, NULL, getmodifiedattributename(),
		NULL, NULL, getipfunc, 0);

	lgba_search(NULL, "(status=*)", NULL, getmodifiedattributename(),
		NULL, NULL, getipfunc, 0);

	lgba_close();
	exit(1);
};

int getipfunc(int num) {
	format("addr=${object}\n", 1);
	syslog(LOG_INFO, "%s got ip: %s ('%s' => 'leasing')", u, getcurrentobjectname(), getformat("${status}"));
	setattr("status", "leasing");
	lgba_close();
	exit(2);
};

int countonlineuser(char *name) {
	FILE *fd;
	struct utmp up;
	int count=0;
	static char *utmpfile=NULL;

	if(!utmpfile) utmpfile=cfgget("utmpfile");
	if(!utmpfile || !(fd=fopen(utmpfile, "r"))) return 0;

	while(fread(&up, sizeof(struct utmp), 1, fd) > 0)
		if(!strncmp(up.ut_name, name, NAS_USERNAME_SIZE)) count++;

	fclose(fd);

	return count;
};
