#include <sys/types.h>
#include <lgba.h>
#include <strfunc.h>
#include <stdio.h>
#include <stdlib.h>
#include <sys/stat.h>
#include <sys/wait.h>
#include <sys/file.h>
#include <time.h>
#include <locale.h>
#include <unistd.h>
#include <syslog.h>
#include "tac_plus.h"
#include "utmp.h"
#include <cost.h>

int ufd;
int slots=0;
off_t size=0;
struct utmp *ufile;

int iterate();
int ifclearable();
int countuser(const char *);
int clearline(struct utmp *);
int getspent(const char *);

typedef struct _ouruser {
	time_t ut_time;
	time_t timespent;
	time_t account;
	time_t strict;
	char username[NAS_USERNAME_SIZE];
	int clearable;
} ouruser;

short int	grefresh=0;

void *ourp;

#ifndef CONFIG
#	define	CONFIG	"/adm/tac/conf/utmpacct.conf"
#endif

#ifndef UTMPCTL_PID_FILE
#	define	UTMPCTL_PID_FILE "/var/run/utmpctl.pid"
#endif

int decusers(ouruser *);

char *account;
slist *prot, *clc;

int main() {
	struct stat sb;
	int les;
	int pf;
	char *s;

	setlocale(LC_ALL, "");

	if(cfgread(CONFIG))
		exit(2);

	if(!(account = cfgget("account")))
		account="timeleft";

	prot=sinit(NULL);
	simport(prot, mkarray(cfgget2("protectedclasses"), 0));
	clc=sinit(NULL);
	simport(clc, mkarray(cfgget2("clearlinefromclasses"), 0));

	if((pf=open(UTMPCTL_PID_FILE, O_WRONLY | O_CREAT, 0644)) == -1) {
		fprintf(stderr, "Error: can't open pid file: %s.\n", UTMPCTL_PID_FILE);
		exit(2);
	};
	if(flock(pf, LOCK_EX | LOCK_NB)) {
		close(pf);
		fprintf(stderr, "Error: previous copy already running.\n");
		exit(2);
	};

	if(!(s=cfgget("utmpfile")) || ((ufd=open(s, O_RDONLY)) == -1) || fstat(ufd, &sb)) {
		fprintf(stderr, "Error: can't open utmp file.\n");
		exit(2);
	};

	if(!(slots=(size=sb.st_size)/sizeof(struct utmp))) {
		fprintf(stderr, "Error: utmp file is empty.\n");
		exit(2);
	};

	ufile=(struct utmp *)malloc(slots * sizeof(struct utmp));

	if(lc_init(NULL)) {
		fprintf(stderr, "Error: libcost init error.\n");
		exit(2);
	};

	if((les=lgba_init(NULL)) || (les=lgba_open())) {
		fprintf(stderr, "Error: LGBA error: %s.\n",
			lgba_err2string(les));
		exit(2);
	}

	ourp=calloc(slots, sizeof(ouruser));

#ifdef AEMON_MODE
		fprintf(stderr, "UTMPctl: All OK. Daemonizing.\n");
		daemon(0, 0);
#endif

	/* saving PID */
	les=strlen(s=genstr(getpid()))+1;
	s[les-1]='\n';
	write(pf, s, les);
	ftruncate(pf, les);
	fsync(pf);

	openlog("utmpctl", LOG_CONS, LOG_LOCAL6);
	syslog(LOG_INFO, "UTMP control facility started: %d.", getpid());

	iterate();

	return 0;
};

#ifndef SLEEP
#	define	SLEEP	60
#endif

/* Refresh accounts at every fifth SLEEP */
#ifndef REFRESH
	#define REFRESH		5
#endif

int iterate() {
	struct stat sb;
	struct utmp *up;
	ouruser *our;
	int slt;
	time_t oldtime, curtime, tmpt;
	time_t ftime=0;

	time(&oldtime);
	fstat(ufd, &sb);

while(1) {
	up=(struct utmp *)ufile;
	our=(ouruser *)ourp;

	lseek(ufd, 0, SEEK_SET);
	read(ufd, ufile, size);

	oldtime=curtime;
	time(&curtime);

	ftime=sb.st_mtime;

	for(slt=0; slt < slots; slt++, up++, our++) {
		if(!*(up->ut_name)) up->ut_time=0;

		/* 1. If utmp and cache are free */
		if(up->ut_time == 0 && our->ut_time == 0) continue;

		/* 2. If user just hanged up */
		if(up->ut_time == 0) {
			decusers(our);
			our->ut_time=0;
			syslog(LOG_INFO, "User %s hanged up; time spent online: %lu.", our->username, our->timespent);
			continue;
		};

		/* 3. If user connected or changed fastly */
		if(up->ut_time > our->ut_time ||
		strncmp(our->username, up->ut_name, NAS_USERNAME_SIZE)) {
			decusers(our);

			*our->username=0;
			our->ut_time=0;
			our->timespent=getspent(up->ut_name);
			our->ut_time = up->ut_time;
			strncpy(our->username, up->ut_name, NAS_USERNAME_SIZE);

			load(our->username);
			tmpt=(our->timespent+=lc_timecost(our->ut_time,
				(time_t)difftime(curtime, our->ut_time),
				"online", getattr2("objectclass")));
			our->account=(time_t)atol(getattr(account));
			our->clearable=ifclearable();
			our->strict=(time_t)atol(getattr("maxonline"));
			if(our->strict && (our->strict < our->account))
				our->account=our->strict;

			syslog(LOG_INFO, "User %s connected; %lu seconds already spent; %s.", our->username, our->timespent, our->clearable?"clearable":"protected");
		} else {
			load(our->username);
			tmpt=lc_timecost(oldtime,
			(time_t)difftime(curtime, oldtime), "online",
			getattr2("objectclass")) * countuser(our->username);

			our->timespent+=tmpt;
		};

		if(our->clearable && our->timespent >= our->account) {

			if(!our->strict) {
				load(our->username);
				our->account=(time_t)atol(getattr(account));
			};

			if(our->timespent >= our->account)
				clearline(up);
		} else {
			syslog(LOG_INFO, "User %s is being served, %ld.",
				our->username, tmpt);

			if(our->clearable && !grefresh && !our->strict) {
				time_t oldvalue=our->account;
				load(our->username);
				our->account=(time_t)atol(getattr(account));
				if(oldvalue != our->account)
				syslog(LOG_INFO,
					"Account of %s refilled from %ld to %ld.",
					our->username, oldvalue, our->account);
			}

		};
	};

	slt=SLEEP;
	while(slt--) {
		usleep(1000000);
		fstat(ufd, &sb);
		if(ftime != sb.st_mtime) break;
	};
	grefresh=(grefresh+1) % REFRESH;
};
};

int countuser(const char *name) {
	int count=0;
	int i;
	ouruser *o=(ouruser *)ourp;
	for(i=0; i < slots; i++, o++) {
		if(!o->ut_time) continue;
		if(!strncasecmp(o->username, name, NAS_USERNAME_SIZE)) count++;
	};
	return count;
};

int ifclearable() {
	char **cls;
	char **ours;

	if(!(cls=getattr2("objectclass"))) {
		return 1;
	};
	
	/* protected classes */
	if(prot && (ours=sarray(prot, 0))) {
		ours--;
		while(*(++ours))
		if(cfind(cls, *ours) != -1) {
			return 0;
		};
	}

	/* clearable classes */
	if(!(ours=sarray(clc, 0)) || !*ours) {
		return 0;
	};

	while(*ours)
	if(cfind(cls, *ours++) != -1) {
		return 1;
	};

	syslog(LOG_INFO, "User without clearable class.\n");
	return 0;
};

int clearline(struct utmp *up) {
	pid_t pid;
	static char *s=NULL;
	syslog(LOG_INFO, "User %s: time limit reached. Clearing.",
		up->ut_name);
	if((pid=fork()) == -1) exit(2);
	if(pid) return 0;

	if(!s) s=cfgget("clearscript");
	if(!s) s=CLEARSCRIPT;

	execl(s, "clear", up->ut_host, up->ut_port, NULL);
	exit(0);
};

int getspent(const char *name) {
	int i;
	ouruser *our=(ouruser *)ourp;

	if(!name || !*name) return 0;

	for(i=0; i<slots; i++, our++)
		if(!strncmp(our->username, name, NAS_USERNAME_SIZE)
			&& our->ut_time)
				return our->timespent;
	return 0;
};


int decusers(ouruser* usr)
{
	int i;
	time_t curtime;
	time_t spent=0;
	ouruser *o=(ouruser*)ourp;

	if(!usr)return 1;
	if(!*(usr->username)) return 1;
	if(!usr->ut_time)return 1;

	time(&curtime);

	load(usr->username);
	spent=lc_timecost(usr->ut_time, (time_t)difftime(curtime, usr->ut_time),
	"online", getattr2("objectclass"));

	for(i=0; i<slots; i++, o++)
		if(!strncmp(o->username, usr->username, NAS_USERNAME_SIZE)) {
			o->timespent-=spent;
			o->account-=spent;
		};
	usr->timespent=spent;
	return 0;
}

