#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <locale.h>
#include <ctype.h>
#include <desire.h>
#include <strfunc.h>

void shac(ctb_data *, int);

void usage();

int main(int ac, char **av) {
	int count=-1;
	ctb_data *data;
	ctb_range range;

	if(ac < 2) usage();

	setlocale(LC_ALL, "");
	if(cfgread(CONFIG))
		exit(2);
	if(ctb_session(NULL, "tacacs", NULL)) exit(2);

	range.either=1;
	range.HDWMY=DAY;
	range.CLP=LAST;

	if(ac > 2) {
		if(!strcmp(av[2], "-h"))	range.HDWMY=HOUR;
		else if(!strcmp(av[2], "-d"))	range.HDWMY=DAY;
		else if(!strcmp(av[2], "-w"))	range.HDWMY=WEEK;
		else if(!strcmp(av[2], "-m"))	range.HDWMY=MONTH;
		else if(!strcmp(av[2], "-y"))	range.HDWMY=YEAR;
		else usage();
	};

	{ char *p;
	p=av[1];
	while(*p) *p++=tolower(*p);
	};

	if(ctb_chname(av[1])) {
		fprintf(stderr, "Can't change name to %s.", av[1]);
		exit(2);
	};

	data=ctb_get(&range, &count);
	if(!data || count < 0) {
		fprintf(stderr, "CTB Library error.\n");
		exit(2);
	};
	if(!count) {
		fprintf(stderr, "No data for this period available.\n");
		exit(2);
	};

	shac(data, count);

	return 0;
}

char *FMT="%s %3d %4d:%02d:%02d  %-9d %-9d %-9d %d\n";

void shac(ctb_data *dt, int count) {
	char *s;
	int cost=0;
	unsigned int tme=0, b_i=0, b_o=0;

	fprintf(stderr, "%-25s %-5s %-8s %-9s %-9s %-9s %s\n", "Date hanged-up", "Op", "Time", "In", "Out", "Cost", "Left");
	while(count--) {
		s=ctime((time_t *)&dt->TS);
		*(strchr(s, '\n'))=0;
		printf(FMT,
			s,
			dt->FL.c.s1c1,
			(int)(dt->U2/ 3600),
			(int)(dt->U2%3600/60),
			dt->U2%60,
			dt->In,
			dt->Out,
			dt->Src,
			dt->Dst
		);
		if(dt->FL.c.s1c1 == 0) cost+=(int)dt->Src;
		tme+=dt->U2;
		b_i+=dt->In;
		b_o+=dt->Out;
		dt++;
	};

	fprintf(stderr, "Overall time: %u:%02u:%02u\n", tme/3600, (tme%3600)/60,tme%60);
	fprintf(stderr, "Overall cost: %d\n", cost);
	if(b_i < 4000)
	fprintf(stderr, "Overall bytes in:  %u\n", b_i);
	else if(b_i < 10000000)
	fprintf(stderr, "Overall bytes in:  %u k.\n", b_i / 1024);
	else
	fprintf(stderr, "Overall bytes in:  %u M.\n", b_i / 1048576 );
	if(b_o < 4000)
	fprintf(stderr, "Overall bytes out: %u\n", b_o);
	else if(b_o < 10000000)
	fprintf(stderr, "Overall bytes out: %u k.\n", b_o / 1024);
	else
	fprintf(stderr, "Overall bytes out: %u M.\n", b_o / 1048576 );

};

void usage() {
	fprintf(stderr, "xac (c) Lev Walkin, 1998-2000.
Usage:
	xac <username> -h
	xac <username> [ -d ]
	xac <username> -w
	xac <username> -m
	xac <username> -y

	Note: month_num's range is 1-12, 0 is 'previous'.
");
	exit(2);
};
